@@ -1,11 +1,15 @@
-__version__ = "4.0.0"
+__version__ = "4.0.1"
 
 
 from .config_resolver import ConfigResolver  # noqa: F401
 from .proc_wrapper import ProcWrapper  # noqa: F401
-from .proc_wrapper_params import CONFIG_MERGE_STRATEGY_SHALLOW  # noqa: F401
-from .proc_wrapper_params import ConfigResolverParams  # noqa: F401
-from .proc_wrapper_params import ProcWrapperParams, make_arg_parser  # noqa: F401
+from .proc_wrapper_params import (  # noqa: F401
+    CONFIG_MERGE_STRATEGY_SHALLOW,
+    DEFAULT_LOG_LEVEL,
+    ConfigResolverParams,
+    ProcWrapperParams,
+    make_arg_parser,
+)
 from .runtime_metadata import (  # noqa: F401
     DefaultRuntimeMetadataFetcher,
     RuntimeMetadata,
