@@ -42,7 +42,11 @@ from urllib.request import Request, urlopen
 from .common_utils import encode_int, string_to_bool
 from .config_resolver import ConfigResolver
 from .proc_wrapper_params import ProcWrapperParams, ProcWrapperParamValidationErrors
-from .runtime_metadata import RuntimeMetadata, RuntimeMetadataFetcher
+from .runtime_metadata import (
+    DefaultRuntimeMetadataFetcher,
+    RuntimeMetadata,
+    RuntimeMetadataFetcher,
+)
 
 _logger = logging.getLogger(__name__)
 _logger.addHandler(logging.NullHandler())
@@ -59,14 +63,15 @@ def _exit_handler(wrapper: "ProcWrapper"):
 def _signal_handler(signum, frame):
     global caught_sigterm
     caught_sigterm = True
-    # This will cause the exit handler to be executed, if it is registered.
     _logger.info(f"Caught signal {signum}, exiting")
 
+    # This will cause the exit handler to be executed, if it is registered.
     # TODO: use different exit code if configured
     sys.exit(0)
 
 
 class ProcWrapper:
+    WRAPPER_FAMILY = "CloudReactor python proc_wrapper"
     VERSION = getattr(sys.modules[__package__], "__version__")
 
     STATUS_RUNNING = "RUNNING"
@@ -106,6 +111,8 @@ class ProcWrapper:
 
     _STATUS_BUFFER_SIZE = 4096
 
+    _STATUS_UPDATE_KEY_LAST_APP_HEARTBEAT_AT = "last_app_heartbeat_at"
+
     def __init__(
         self,
         params: Optional[ProcWrapperParams] = None,
@@ -113,6 +120,7 @@ class ProcWrapper:
         config_resolver: Optional[ConfigResolver] = None,
         env_override: Optional[Mapping[str, Any]] = None,
         override_params_from_env: bool = True,
+        runtime_context: Optional[Any] = None,
     ) -> None:
         _logger.info("Creating ProcWrapper instance ...")
 
@@ -144,25 +152,30 @@ class ProcWrapper:
         self._status_buffer: Optional[bytearray] = None
         self._status_message_so_far: Optional[bytearray] = None
         self.last_update_sent_at: Optional[float] = None
+        self.last_app_heartbeat_at: Optional[datetime] = None
 
         self.rollbar_retries_exhausted = False
         self.exit_handler_installed = False
         self.in_pytest = False
 
         self.runtime_metadata_fetcher: RuntimeMetadataFetcher = (
-            runtime_metadata_fetcher or RuntimeMetadataFetcher()
+            runtime_metadata_fetcher or DefaultRuntimeMetadataFetcher()
         )
 
+        self.runtime_context = runtime_context
+
         runtime_metadata: Optional[RuntimeMetadata] = None
         try:
-            runtime_metadata = runtime_metadata_fetcher.fetch(env=self.env)
+            runtime_metadata = self.runtime_metadata_fetcher.fetch(
+                env=self.env, context=self.runtime_context
+            )
         except Exception:
             _logger.exception("Failed to fetch runtime metadata")
 
         if params:
             self.params = params
         else:
-            self.params = ProcWrapperParams()
+            self.params = ProcWrapperParams(override_from_env=False)
             self.params.override_resolver_params_from_env(env=self.env)
 
         if config_resolver:
@@ -212,10 +225,17 @@ class ProcWrapper:
 
         return None
 
-    def log_configuration(self) -> None:
-        _logger.debug(f"Wrapper version = {ProcWrapper.VERSION}")
+    def log_configuration(self, initial: bool = False) -> None:
+        if initial:
+            _logger.info(f"Wrapper version = {ProcWrapper.VERSION}")
+
         self.params.log_configuration()
 
+        if self.param_errors:
+            self.param_errors.log()
+        else:
+            _logger.warning("No validated parameters?!")
+
     def print_final_status(
         self,
         exit_code: Optional[int],
@@ -256,11 +276,11 @@ class ProcWrapper:
 
         if first_attempt_started_at is not None:
             total_duration = now_ts - first_attempt_started_at
-            msg += f", for a total duration of {round(total_duration)} seconds"
+            msg += f", for a total duration of {round(total_duration)} second(s)"
 
         msg += "."
 
-        print(msg)
+        _logger.info(msg)
 
     @property
     def max_execution_attempts(self) -> float:
@@ -269,6 +289,33 @@ class ProcWrapper:
 
         return self.params.process_max_retries + 1
 
+    def update_status(
+        self,
+        success_count: Optional[int] = None,
+        error_count: Optional[int] = None,
+        skipped_count: Optional[int] = None,
+        expected_count: Optional[int] = None,
+        last_status_message: Optional[str] = None,
+        extra_status_props: Optional[Dict[str, Any]] = None,
+        last_app_heartbeat_at: Optional[datetime] = None,
+    ) -> bool:
+        """
+        Update the status of the process. Send to the Task management
+        server if the last status was sent more than status_update_interval
+        seconds ago. Return true if the update was actually sent, instead of
+        buffered.
+        """
+        return self._update_status(
+            success_count=success_count,
+            error_count=error_count,
+            skipped_count=skipped_count,
+            expected_count=expected_count,
+            last_status_message=last_status_message,
+            extra_status_props=extra_status_props,
+            is_app_update=True,
+            last_app_heartbeat_at=last_app_heartbeat_at,
+        )
+
     def _create_or_update_task_execution(self) -> None:
         """
         Make a request to the API server to create a Task Execution
@@ -289,7 +336,7 @@ class ProcWrapper:
         runtime_metadata: Optional[RuntimeMetadata] = None
         if self.params.send_runtime_metadata:
             runtime_metadata = self.runtime_metadata_fetcher.fetch(
-                env=self.resolved_env
+                env=self.resolved_env, context=self.runtime_context
             )
 
         status = ProcWrapper.STATUS_RUNNING
@@ -332,6 +379,7 @@ class ProcWrapper:
                 "prevent_offline_execution": self.params.prevent_offline_execution,
                 "process_termination_grace_period_seconds": self.params.process_termination_grace_period,
                 "wrapper_log_level": logging.getLevelName(_logger.getEffectiveLevel()),
+                "wrapper_family": ProcWrapper.WRAPPER_FAMILY,
                 "wrapper_version": ProcWrapper.VERSION,
                 "api_error_timeout_seconds": encode_int(
                     self.params.api_error_timeout, empty_value=-1
@@ -481,6 +529,7 @@ class ProcWrapper:
 
                     if not self.task_execution_uuid:
                         self.task_execution_uuid = response_dict.get("uuid")
+                        self._reconfigure_logger()
 
                     self.task_uuid = self.task_uuid or response_dict["task"]["uuid"]
                     self.task_name = self.task_name or response_dict["task"]["name"]
@@ -494,10 +543,11 @@ class ProcWrapper:
 
             _logger.info("Not preventing offline execution, so continuing")
 
-    def update_status(
+    def _update_status(
         self,
         failed_attempts: Optional[int] = None,
         timed_out_attempts: Optional[int] = None,
+        exit_code: Optional[int] = None,
         pid: Optional[int] = None,
         success_count: Optional[int] = None,
         error_count: Optional[int] = None,
@@ -505,14 +555,21 @@ class ProcWrapper:
         expected_count: Optional[int] = None,
         last_status_message: Optional[str] = None,
         extra_status_props: Optional[Dict[str, Any]] = None,
-    ) -> None:
-        """Update the status of the process. Send to the process management
-        server if the last status was sent more than status_update_interval
-        seconds ago.
-        """
+        is_app_update: bool = False,
+        last_app_heartbeat_at: Optional[datetime] = None,
+    ) -> bool:
+        _logger.debug(f"_update_status(), is_app_update = {is_app_update}")
+
+        if is_app_update:
+            self.last_app_heartbeat_at = last_app_heartbeat_at or datetime.now()
 
         if self.offline_mode:
-            return
+            return False
+
+        if is_app_update:
+            self.status_dict[
+                self._STATUS_UPDATE_KEY_LAST_APP_HEARTBEAT_AT
+            ] = self.last_app_heartbeat_at
 
         if success_count is not None:
             self.status_dict["success_count"] = int(success_count)
@@ -532,21 +589,27 @@ class ProcWrapper:
         if extra_status_props:
             self.status_dict.update(extra_status_props)
 
-        should_send = self.is_status_update_due()
-
         # These are important updates that should be sent no matter what.
         if not (
-            (failed_attempts is None) and (timed_out_attempts is None) and (pid is None)
-        ) and (last_status_message is None):
+            (failed_attempts is None)
+            and (timed_out_attempts is None)
+            and (pid is None)
+            and (exit_code is None)
+        ):
             should_send = True
+        else:
+            should_send = self.is_status_update_due()
 
         if should_send:
-            self.send_update(
+            return self.send_update(
                 failed_attempts=failed_attempts,
                 timed_out_attempts=timed_out_attempts,
+                exit_code=exit_code,
                 pid=pid,
             )
 
+        return False
+
     def send_completion(
         self,
         status: str,
@@ -564,7 +627,7 @@ class ProcWrapper:
             self._exit_or_raise(self._EXIT_CODE_GENERIC_ERROR)
             return
 
-        if exit_code is None:
+        if (exit_code is None) and (not self.params.embedded_mode):
             if status == self.STATUS_SUCCEEDED:
                 exit_code = self._EXIT_CODE_SUCCESS
             else:
@@ -576,6 +639,7 @@ class ProcWrapper:
             timed_out_attempts=timed_out_attempts,
             exit_code=exit_code,
             pid=pid,
+            finished_at=datetime.now(),
         )
 
     def managed_call(self, fun, data: Any = None):
@@ -601,11 +665,12 @@ class ProcWrapper:
             return
 
         self._reload_params()
+        self.log_configuration(initial=True)
         self._setup_task_execution()
 
         rv = None
         success = False
-        saved_ex = None
+        saved_ex: Exception = RuntimeError()
         while self.attempt_count < self.max_execution_attempts:
             self.attempt_count += 1
 
@@ -615,6 +680,11 @@ class ProcWrapper:
 
             try:
                 rv = fun(self, data, self.resolved_config)
+
+                _logger.info("Managed function succeeded")
+
+                # TODO: log and send return value back to server
+
                 success = True
             except Exception as ex:
                 saved_ex = ex
@@ -622,7 +692,9 @@ class ProcWrapper:
                 _logger.exception("Managed function failed")
 
                 if self.attempt_count < self.max_execution_attempts:
-                    self.send_update(failed_attempts=self.failed_count)
+                    self._update_status(failed_attempts=self.failed_count)
+
+                    # TODO: check timeout
 
                     if self.params.process_retry_delay:
                         _logger.debug("Sleeping after managed function failed ...")
@@ -631,12 +703,18 @@ class ProcWrapper:
 
                     if self.params.config_ttl is not None:
                         self._reload_params()
+                        self.log_configuration()
 
             if success:
-                self.send_completion(
-                    status=ProcWrapper.STATUS_SUCCEEDED,
-                    failed_attempts=self.failed_count,
-                )
+                try:
+                    self.send_completion(
+                        status=ProcWrapper.STATUS_SUCCEEDED,
+                    )
+                except Exception:
+                    _logger.warning(
+                        "Failed to send completion to the API server", exc_info=True
+                    )
+
                 return rv
 
         self.send_completion(
@@ -655,7 +733,9 @@ class ProcWrapper:
             want_env=True, want_config=self.params.embedded_mode
         )
 
-        runtime_metadata = self.runtime_metadata_fetcher.fetch(env=self.resolved_env)
+        runtime_metadata = self.runtime_metadata_fetcher.fetch(
+            env=self.resolved_env, context=self.runtime_context
+        )
 
         # In case API key(s) change
         self.params.override_proc_wrapper_params_from_env(
@@ -667,15 +747,6 @@ class ProcWrapper:
         )
 
     def _setup_task_execution(self) -> bool:
-        self._reload_params()
-
-        self.log_configuration()
-
-        if self.param_errors:
-            self.param_errors.log()
-        else:
-            _logger.warning("No validated parameters?!")
-
         self.task_uuid = self.params.task_uuid
         self.task_name = self.params.task_name
         self.task_execution_uuid = self.params.task_execution_uuid
@@ -686,6 +757,9 @@ class ProcWrapper:
         self.attempt_count = 0
         self.timeout_count = 0
 
+        if self.task_execution_uuid:
+            self._reconfigure_logger()
+
         if self.offline_mode:
             _logger.info("Starting in offline mode ...")
         else:
@@ -707,22 +781,29 @@ class ProcWrapper:
 
         return True
 
-    def run(self) -> None:
+    def run(self) -> int:
         """
         Run the wrapped process, first requesting to start, then executing
         the process command line, retying for errors and timeouts.
+        The return value is the exit code, but won't be actually returned
+        except in testing.
         """
         self._ensure_non_embedded_mode()
 
+        self._reload_params()
+
+        self.log_configuration(initial=True)
+
+        if self.param_errors and not self.param_errors.can_start_process():
+            _logger.error("Exiting due to process configuration error")
+            return self._exit_or_raise(self._EXIT_CODE_CONFIGURATION_ERROR)
+
+        command, shell = self.params.resolve_command_and_shell_flag()
+
         should_run = self._setup_task_execution()
 
-        if not self.params.command:
-            _logger.error("No command found in wrapped mode")
-            self._exit_or_raise(self._EXIT_CODE_CONFIGURATION_ERROR)
-            return
-
         if not should_run:
-            self._exit_or_raise(self._EXIT_CODE_GENERIC_ERROR)
+            return self._exit_or_raise(self._EXIT_CODE_GENERIC_ERROR)
 
         if (not self.offline_mode) and self.params.enable_status_update_listener:
             self._status_buffer = bytearray(self._STATUS_BUFFER_SIZE)
@@ -763,10 +844,6 @@ class ProcWrapper:
                     current_time + self.params.process_check_interval
                 )
 
-            _logger.debug(
-                f"command = {self.params.command}, cwd={self.params.work_dir}"
-            )
-
             if self.params.log_secrets:
                 _logger.debug(f"process_env={process_env}")
 
@@ -777,22 +854,27 @@ class ProcWrapper:
             # Set the session ID so we can kill the process as a group, so we kill
             # all subprocesses. See https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true
             # On Windows, os does not have setsid function.
-            # TODO: allow non-shell mode
+            preexec_fn = (
+                getattr(os, "setsid", None)
+                if self.params.process_group_termination
+                else None
+            )
+
             self.process = Popen(
-                " ".join(self.params.command),
-                shell=True,
+                command,
+                shell=shell,
                 stdout=None,
                 stderr=None,
                 env=process_env,
                 cwd=self.params.work_dir,
-                preexec_fn=getattr(os, "setsid", None),
+                preexec_fn=preexec_fn,
             )
 
             pid = self.process.pid
             _logger.info(f"pid = {pid}")
 
             if pid and self.params.send_pid:
-                self.send_update(pid=pid)
+                self._update_status(pid=pid)
 
             done_polling = False
             while not done_polling:
@@ -810,8 +892,8 @@ class ProcWrapper:
                         self.timeout_count += 1
 
                         if self.attempt_count < self.max_execution_attempts:
-                            self.send_update(
-                                failed_attempts=self.failed_count,
+                            # FIXME: clear exit code
+                            self._update_status(
                                 timed_out_attempts=self.timeout_count,
                             )
 
@@ -869,8 +951,7 @@ class ProcWrapper:
                             first_attempt_started_at=first_attempt_started_at,
                             latest_attempt_started_at=latest_attempt_started_at,
                         )
-                        self._exit_or_raise(0)
-                        return
+                        return self._exit_or_raise(0)
 
                     self.failed_count += 1
 
@@ -880,10 +961,9 @@ class ProcWrapper:
                             first_attempt_started_at=first_attempt_started_at,
                             latest_attempt_started_at=latest_attempt_started_at,
                         )
-                        self._exit_or_raise(exit_code)
-                        return
+                        return self._exit_or_raise(exit_code)
 
-                    self.send_update(
+                    self._update_status(
                         failed_attempts=self.failed_count, exit_code=exit_code
                     )
 
@@ -896,7 +976,9 @@ class ProcWrapper:
 
                     if self.params.config_ttl is not None:
                         self._reload_params()
+                        self.log_configuration()
                         process_env = self.make_process_env()
+                        command, shell = self.params.resolve_command_and_shell_flag()
 
         if self.attempt_count >= self.max_execution_attempts:
             self.print_final_status(
@@ -905,11 +987,15 @@ class ProcWrapper:
                 latest_attempt_started_at=latest_attempt_started_at,
             )
 
-            self._exit_or_raise(self._EXIT_CODE_GENERIC_ERROR)
+            return self._exit_or_raise(self._EXIT_CODE_GENERIC_ERROR)
         else:
-            self.send_update(failed_attempts=self.attempt_count)
+            self.send_update(
+                failed_attempts=self.attempt_count,
+                timed_out_attempts=self.timeout_count,
+            )
 
         # Exit handler will send update to API server
+        return 0
 
     def handle_exit(self) -> None:
         _logger.debug(f"Exit handler, Task Execution UUID = {self.task_execution_uuid}")
@@ -929,7 +1015,7 @@ class ProcWrapper:
 
         exit_code = None
         pid = None
-        notification_required = True
+        notification_required = not self.offline_mode
 
         try:
             if self.process:
@@ -968,7 +1054,7 @@ class ProcWrapper:
             if self.api_server_retries_exhausted:
                 error_message = "API Server not responding properly"
             elif not self.task_execution_uuid:
-                error_message = "Task Execution ID not assigned"
+                error_message = "Task Execution UUID not assigned"
 
             self._report_error(error_message, self.last_api_request_data)
 
@@ -988,9 +1074,12 @@ class ProcWrapper:
             )
             return exit_code
 
-        _logger.warning("Sending SIGTERM ...")
+        _logger.info("Sending SIGTERM ...")
 
-        os.killpg(os.getpgid(pid), signal.SIGTERM)
+        if self.params.process_group_termination:
+            os.killpg(os.getpgid(pid), signal.SIGTERM)
+        else:
+            os.kill(pid, signal.SIGTERM)
 
         try:
             self.process.communicate(
@@ -1001,7 +1090,11 @@ class ProcWrapper:
         except TimeoutExpired:
             _logger.info("Timeout after SIGTERM expired, killing with SIGKILL ...")
             try:
-                os.killpg(os.getpgid(pid), signal.SIGKILL)
+                if self.params.process_group_termination:
+                    os.killpg(os.getpgid(pid), signal.SIGKILL)
+                else:
+                    os.kill(pid, signal.SIGKILL)
+
                 self.process.communicate()
             except Exception:
                 _logger.exception(f"Could not kill process with pid {pid}")
@@ -1097,11 +1190,11 @@ class ProcWrapper:
                     self.send_update()
             except json.JSONDecodeError:
                 _logger.debug(
-                    "Error decoding JSON, this can happen due to missing out of order messages"
+                    "Error decoding JSON, this can happen due to missing or out of order messages"
                 )
         except UnicodeDecodeError:
             _logger.debug(
-                "Error decoding message as UTF-8, this can happen due to missing out of order messages"
+                "Error decoding message as UTF-8, this can happen due to missing or out of order messages"
             )
         finally:
             self._status_message_so_far.clear()
@@ -1117,7 +1210,7 @@ class ProcWrapper:
 
     def _make_headers(self) -> Dict[str, str]:
         headers = {
-            "Authorization": f"Token {self.params.api_key}",
+            "Authorization": f"Bearer {self.params.api_key}",
             "Content-Type": "application/json",
             "Accept": "application/json",
         }
@@ -1135,27 +1228,27 @@ class ProcWrapper:
         timed_out_attempts: Optional[int] = None,
         exit_code: Optional[int] = None,
         pid: Optional[int] = None,
+        finished_at: Optional[datetime] = None,
         extra_props: Optional[Mapping[str, Any]] = None,
-    ) -> None:
+    ) -> bool:
         """
-        Send an update to the API server. If status is omitted, it will
-        default to STATUS_RUNNING.
+        Send an update to the API server immediately. If status is omitted, it
+        will default to STATUS_RUNNING.
 
         The caller may want to coalesce multiple updates together in order to
         save bandwidth / API usage.
 
-        This method is meant to be called directly in the callback function
-        when running in embedded mode.
+        This method is NOT meant to be called directly.
         """
 
         if self.offline_mode:
-            return
+            return False
 
         if not self.task_execution_uuid:
             _logger.debug(
                 "_send_update() skipping update since no Task Execution UUID was available"
             )
-            return
+            return False
 
         if status is None:
             status = self.STATUS_RUNNING
@@ -1164,10 +1257,10 @@ class ProcWrapper:
 
         body.update(self.status_dict)
 
-        if failed_attempts is not None:
+        if failed_attempts:
             body["failed_attempts"] = failed_attempts
 
-        if timed_out_attempts is not None:
+        if timed_out_attempts:
             body["timed_out_attempts"] = timed_out_attempts
 
         if exit_code is not None:
@@ -1176,10 +1269,26 @@ class ProcWrapper:
         if pid is not None:
             body["pid"] = pid
 
+        if finished_at:
+            body["finished_at"] = finished_at.isoformat()
+
         if extra_props:
             body["other_runtime_metadata"] = extra_props
 
-        url = f"{self.params.api_base_url}/api/v1/task_executions/{quote_plus(str(self.task_execution_uuid))}/"
+        unsent_last_app_heartbeat_at = self.status_dict.get(
+            self._STATUS_UPDATE_KEY_LAST_APP_HEARTBEAT_AT
+        )
+
+        if unsent_last_app_heartbeat_at:
+            body[
+                self._STATUS_UPDATE_KEY_LAST_APP_HEARTBEAT_AT
+            ] = unsent_last_app_heartbeat_at.isoformat()
+
+        url = (
+            f"{self.params.api_base_url}/api/v1/task_executions/"
+            + quote_plus(str(self.task_execution_uuid))
+            + "/?content=false"
+        )
         headers = self._make_headers()
         text_data = json.dumps(body)
         data = text_data.encode("utf-8")
@@ -1191,12 +1300,15 @@ class ProcWrapper:
         f = self._send_api_request(req, is_final_update=is_final_update)
         if f is None:
             _logger.debug("Update request failed non-fatally")
-        else:
-            with f:
-                _logger.debug("Update sent successfully.")
-                self.last_update_sent_at = time.time()
-                self.status_dict = {}
-                self.reported_final_status = is_final_update
+            return False
+
+        with f:
+            _logger.debug("Update sent successfully.")
+            self.last_update_sent_at = time.time()
+            self.status_dict = {}
+            self.reported_final_status = is_final_update
+
+        return True
 
     def _compute_successful_request_deadline(
         self,
@@ -1408,7 +1520,31 @@ class ProcWrapper:
 
         return self.api_server_retries_exhausted
 
-    def _exit_or_raise(self, exit_code) -> None:
+    def _reconfigure_logger(self) -> None:
+        format = "PROC_WRAPPER: "
+
+        if self.task_execution_uuid:
+            format += f"[{self.task_execution_uuid}] "
+
+        if self.params.include_timestamps_in_log:
+            format += "%(asctime)s "
+
+        format += "%(levelname)s: %(message)s"
+
+        formatter = logging.Formatter(format)
+        logger_to_configure = _logger
+
+        if not self.params.embedded_mode:
+            # Root logger
+            logger_to_configure = logging.getLogger()
+
+        for handler in logger_to_configure.handlers:
+            try:
+                handler.setFormatter(formatter)
+            except Exception:
+                _logger.warning("Can't set formatter on logger")
+
+    def _exit_or_raise(self, exit_code: int) -> int:
         if self.params.embedded_mode:
             raise RuntimeError(
                 f"Raising an error in embedded mode, exit code {exit_code}"
@@ -1424,16 +1560,13 @@ class ProcWrapper:
 
             sys.exit(exit_code)
 
-    def _report_error(self, message: str, data: Optional[Dict[str, Any]]) -> int:
-        num_sinks_successful = 0
+        return exit_code
+
+    def _report_error(self, message: str, data: Optional[Dict[str, Any]]) -> None:
+        _logger.error(message)
+
         if self.params.rollbar_access_token:
-            if self._send_rollbar_error(message, data):
-                num_sinks_successful += 1
-
-        if num_sinks_successful == 0:
-            _logger.info("Can't notify any valid error sink!")
-
-        return num_sinks_successful
+            self._send_rollbar_error(message, data)
 
     def _send_rollbar_error(self, message: str, data=None, level="error") -> bool:
         if not self.params.rollbar_access_token:
